# TODO 可以把软件改成运行中加载配置文件的方式
#   例如如果有 "config.ini" 文件则读取
#   否则使用该文件的默认配置

TYPE_ITEMS = ['印刷体', '手写体', '印章', '身份证', '表格', '其它证件类', '其它']
KV_ITEMS = ['key', 'value']

SROIE_KV_ITEMS = ['other', 'key', 'value']

SROIE_CLASS_ITEMS = [
    'other',
    'company',
    'address',
    'date',
    'total'
]

# 一种精简的配置方法，需要使用_py2ls转成规范的配置结构
COCO = [
    ['id', 1, 'int'],
    ['label', 1, 'str'],
    ['category_id', 1, 'int'],
    ['content_type', 1, 'str', tuple(TYPE_ITEMS)],
    # ['content_class', 1, 'str', tuple(CLASS_ITEMS)],
    ['content_kv', 1, 'str', tuple(KV_ITEMS)],
    ['sroie_class', 1, 'str', tuple(SROIE_CLASS_ITEMS)],
    ['sroie_kv', 1, 'str', tuple(SROIE_KV_ITEMS)],
    ['bbox', 0],
    ['area', 0],
    ['image_id', 0],
    ['segmentation', 0],
    ['iscrowd', 0],
    ['points', 0, 'list'],
    ['shape_color', 0, 'list'],
    ['line_color', 0, 'list'],
    ['vertex_color', 0, 'list'],
]


def _py2ls(ls):
    """
    :param ls: 原始手动录入的数据格式
    :return: dict[attr, Item]
    """
    res = []
    for x in ls:
        # 1 补长对齐
        if len(x) < 4:
            x += [None] * (4 - len(x))
        # 2 设置属性值
        d = {'key': x[0], 'show': x[1], 'type': x[2], 'items': x[3]}
        if isinstance(x[3], list):
            d['editable'] = 1
        res.append(d)

    return res


COCO = _py2ls(COCO)



import re

from qtpy import QT_VERSION
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
from qtpy.QtWidgets import QWidgetItem
from xllabelme.logger import logger
import xllabelme.utils
from xllabelme.config.itemcfg import *
from xllabelme.shape import ShapeExt

from pyxllib.gui.qt import get_input_widget, QHLine

QT5 = QT_VERSION[0] == "5"


# TODO(unknown):
# - Calculate optimal position so as not to go out of screen area.


class LabelQLineEdit(QtWidgets.QLineEdit):
    def setListWidget(self, list_widget):
        self.list_widget = list_widget

    def keyPressEvent(self, e):
        if e.key() in [QtCore.Qt.Key_Up, QtCore.Qt.Key_Down]:
            self.list_widget.keyPressEvent(e)
        else:
            super(LabelQLineEdit, self).keyPressEvent(e)


class DefaultLabelDialog(QtWidgets.QDialog):
    def __init__(
            self,
            text="Enter object label",
            parent=None,
            labels=None,
            sort_labels=True,
            show_text_field=True,
            completion="startswith",
            fit_to_content=None,
            flags=None,
    ):
        if fit_to_content is None:
            fit_to_content = {"row": False, "column": True}
        self._fit_to_content = fit_to_content

        super(DefaultLabelDialog, self).__init__(parent)
        self.edit = LabelQLineEdit()
        self.edit.setPlaceholderText(text)
        self.edit.setValidator(xllabelme.utils.labelValidator())
        self.edit.editingFinished.connect(self.postProcess)
        if flags:
            self.edit.textChanged.connect(self.updateFlags)
        self.edit_group_id = QtWidgets.QLineEdit()
        self.edit_group_id.setPlaceholderText("Group ID")
        self.edit_group_id.setValidator(
            QtGui.QRegExpValidator(QtCore.QRegExp(r"\d*"), None)
        )
        layout = QtWidgets.QVBoxLayout()
        if show_text_field:
            layout_edit = QtWidgets.QHBoxLayout()
            layout_edit.addWidget(self.edit, 6)
            layout_edit.addWidget(self.edit_group_id, 2)
            layout.addLayout(layout_edit)
        # buttons
        self.buttonBox = bb = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            QtCore.Qt.Horizontal,
            self,
        )
        bb.button(bb.Ok).setIcon(xllabelme.utils.newIcon("done"))
        bb.button(bb.Cancel).setIcon(xllabelme.utils.newIcon("undo"))
        bb.accepted.connect(self.validate)
        bb.rejected.connect(self.reject)
        layout.addWidget(bb)
        # label_list
        self.labelList = QtWidgets.QListWidget()
        if self._fit_to_content["row"]:
            self.labelList.setHorizontalScrollBarPolicy(
                QtCore.Qt.ScrollBarAlwaysOff
            )
        if self._fit_to_content["column"]:
            self.labelList.setVerticalScrollBarPolicy(
                QtCore.Qt.ScrollBarAlwaysOff
            )
        self._sort_labels = sort_labels
        if labels:
            self.labelList.addItems(labels)
        if self._sort_labels:
            self.labelList.sortItems()
        else:
            self.labelList.setDragDropMode(
                QtWidgets.QAbstractItemView.InternalMove
            )
        self.labelList.currentItemChanged.connect(self.labelSelected)
        self.labelList.itemDoubleClicked.connect(self.labelDoubleClicked)
        self.edit.setListWidget(self.labelList)
        layout.addWidget(self.labelList)
        # label_flags
        if flags is None:
            flags = {}
        self._flags = flags
        self.flagsLayout = QtWidgets.QVBoxLayout()
        self.resetFlags()
        layout.addItem(self.flagsLayout)
        self.edit.textChanged.connect(self.updateFlags)
        self.setLayout(layout)
        # completion
        completer = QtWidgets.QCompleter()
        if not QT5 and completion != "startswith":
            logger.warn(
                "completion other than 'startswith' is only "
                "supported with Qt5. Using 'startswith'"
            )
            completion = "startswith"
        if completion == "startswith":
            completer.setCompletionMode(QtWidgets.QCompleter.InlineCompletion)
            # Default settings.
            # completer.setFilterMode(QtCore.Qt.MatchStartsWith)
        elif completion == "contains":
            completer.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
            completer.setFilterMode(QtCore.Qt.MatchContains)
        else:
            raise ValueError("Unsupported completion: {}".format(completion))
        completer.setModel(self.labelList.model())
        self.edit.setCompleter(completer)

    def addLabelHistory(self, label):
        if self.labelList.findItems(label, QtCore.Qt.MatchExactly):
            return
        self.labelList.addItem(label)
        if self._sort_labels:
            self.labelList.sortItems()

    def labelSelected(self, item):
        self.edit.setText(item.text())

    def validate(self):
        text = self.edit.text()
        if hasattr(text, "strip"):
            text = text.strip()
        else:
            text = text.trimmed()
        if text:
            self.accept()

    def labelDoubleClicked(self, item):
        self.validate()

    def postProcess(self):
        text = self.edit.text()
        if hasattr(text, "strip"):
            text = text.strip()
        else:
            text = text.trimmed()
        self.edit.setText(text)

    def updateFlags(self, label_new):
        # keep state of shared flags
        flags_old = self.getFlags()

        flags_new = {}
        for pattern, keys in self._flags.items():
            if re.match(pattern, label_new):
                for key in keys:
                    flags_new[key] = flags_old.get(key, False)
        self.setFlags(flags_new)

    def deleteFlags(self):
        for i in reversed(range(self.flagsLayout.count())):
            item = self.flagsLayout.itemAt(i).widget()
            self.flagsLayout.removeWidget(item)
            item.setParent(None)

    def resetFlags(self, label=""):
        flags = {}
        for pattern, keys in self._flags.items():
            if re.match(pattern, label):
                for key in keys:
                    flags[key] = False
        self.setFlags(flags)

    def setFlags(self, flags):
        self.deleteFlags()
        for key in flags:
            item = QtWidgets.QCheckBox(key, self)
            item.setChecked(flags[key])
            self.flagsLayout.addWidget(item)
            item.show()

    def getFlags(self):
        flags = {}
        for i in range(self.flagsLayout.count()):
            item = self.flagsLayout.itemAt(i).widget()
            flags[item.text()] = item.isChecked()
        return flags

    def getGroupId(self):
        group_id = self.edit_group_id.text()
        if group_id:
            return int(group_id)
        return None

    def popUp(self, text=None, move=True, flags=None, group_id=None):
        if self._fit_to_content["row"]:
            self.labelList.setMinimumHeight(
                self.labelList.sizeHintForRow(0) * self.labelList.count() + 2
            )
        if self._fit_to_content["column"]:
            self.labelList.setMinimumWidth(
                self.labelList.sizeHintForColumn(0) + 2
            )
        # if text is None, the previous label in self.edit is kept
        if text is None:
            text = self.edit.text()
        if flags:
            self.setFlags(flags)
        else:
            self.resetFlags(text)
        self.edit.setText(text)
        self.edit.setSelection(0, len(text))
        if group_id is None:
            self.edit_group_id.clear()
        else:
            self.edit_group_id.setText(str(group_id))
        items = self.labelList.findItems(text, QtCore.Qt.MatchFixedString)
        if items:
            if len(items) != 1:
                logger.warning("Label list has duplicate '{}'".format(text))
            self.labelList.setCurrentItem(items[0])
            row = self.labelList.row(items[0])
            self.edit.completer().setCurrentRow(row)
        self.edit.setFocus(QtCore.Qt.PopupFocusReason)
        if move:
            self.move(QtGui.QCursor.pos())
        if self.exec_():
            return self.edit.text(), self.getFlags(), self.getGroupId()
        else:
            return None, None, None


class LabelDialogExt(DefaultLabelDialog):
    def __init__(
            self,
            text="Enter object label",
            parent=None,
            labels=None,
            sort_labels=True,
            show_text_field=True,
            completion="startswith",
            fit_to_content=None,
            flags=None,
    ):
        if fit_to_content is None:
            fit_to_content = {"row": False, "column": True}
        self._fit_to_content = fit_to_content

        super(DefaultLabelDialog, self).__init__(parent)
        self.setWindowTitle(f'view/edit label[*]')
        self.edit = LabelQLineEdit()
        self.edit.setPlaceholderText(text)
        self.edit.setValidator(xllabelme.utils.labelValidator())
        self.edit.editingFinished.connect(self.postProcess)
        if flags:
            self.edit.textChanged.connect(self.updateFlags)
        self.edit_group_id = QtWidgets.QLineEdit()
        self.edit_group_id.setPlaceholderText("Group ID")
        self.edit_group_id.setValidator(
            QtGui.QRegExpValidator(QtCore.QRegExp(r"\d*"), None)
        )
        # layout = QtWidgets.QVBoxLayout()
        layout = QtWidgets.QFormLayout()
        if show_text_field:
            layout_edit = QtWidgets.QHBoxLayout()
            layout_edit.addWidget(self.edit, 6)
            layout_edit.addWidget(self.edit_group_id, 2)
            layout_edit.setProperty('labelme origin layout', True)
            layout.addRow(layout_edit)
        # buttons
        self.buttonBox = bb = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            QtCore.Qt.Horizontal,
            self,
        )
        bb.button(bb.Ok).setIcon(xllabelme.utils.newIcon("done"))
        bb.button(bb.Cancel).setIcon(xllabelme.utils.newIcon("undo"))
        bb.accepted.connect(self.validate)
        bb.rejected.connect(self.reject)
        bb.setProperty('labelme origin layout', True)
        layout.addRow(bb)
        # label_list
        self.labelList = QtWidgets.QListWidget(self)
        # self.labelList.destroyed.connect(lambda: print('labelList被删除'))
        if self._fit_to_content["row"]:
            self.labelList.setHorizontalScrollBarPolicy(
                QtCore.Qt.ScrollBarAlwaysOff
            )
        if self._fit_to_content["column"]:
            self.labelList.setVerticalScrollBarPolicy(
                QtCore.Qt.ScrollBarAlwaysOff
            )
        self._sort_labels = sort_labels
        if labels:
            self.labelList.addItems(labels)
        if self._sort_labels:
            self.labelList.sortItems()
        else:
            self.labelList.setDragDropMode(
                QtWidgets.QAbstractItemView.InternalMove
            )
        self.labelList.currentItemChanged.connect(self.labelSelected)
        self.labelList.itemDoubleClicked.connect(self.labelDoubleClicked)
        self.edit.setListWidget(self.labelList)
        self.labelList.setProperty('labelme origin layout', True)
        layout.addRow(self.labelList)
        # label_flags
        if flags is None:
            flags = {}
        self._flags = flags
        self.flagsLayout = QtWidgets.QVBoxLayout()
        self.resetFlags()
        self.flagsLayout.setProperty('labelme origin layout', True)
        layout.addRow(self.flagsLayout)
        self.edit.textChanged.connect(self.updateFlags)
        self.setLayout(layout)
        # completion
        completer = QtWidgets.QCompleter()
        if not QT5 and completion != "startswith":
            logger.warn(
                "completion other than 'startswith' is only "
                "supported with Qt5. Using 'startswith'"
            )
            completion = "startswith"
        if completion == "startswith":
            completer.setCompletionMode(QtWidgets.QCompleter.InlineCompletion)
            # Default settings.
            # completer.setFilterMode(QtCore.Qt.MatchStartsWith)
        elif completion == "contains":
            completer.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
            completer.setFilterMode(QtCore.Qt.MatchContains)
        else:
            raise ValueError("Unsupported completion: {}".format(completion))
        completer.setModel(self.labelList.model())
        self.edit.setCompleter(completer)
        self.geo = None  # 存储相对桌面屏幕的坐标位置

    def exec_(self):
        if self.geo:
            self.setGeometry(self.geo)
        res = super().exec_()
        self.geo = self.geometry()
        return res

    def popUp(self, text=None, move=True, flags=None, group_id=None):
        self.initLayout()

        if self._fit_to_content["row"]:
            self.labelList.setMinimumHeight(
                self.labelList.sizeHintForRow(0) * self.labelList.count() + 2
            )
        if self._fit_to_content["column"]:
            self.labelList.setMinimumWidth(
                self.labelList.sizeHintForColumn(0) + 2
            )
        # if text is None, the previous label in self.edit is kept
        if text is None:
            text = self.edit.text()
        if flags:
            self.setFlags(flags)
        else:
            self.resetFlags(text)
        self.edit.setText(text)
        self.edit.setSelection(0, len(text))
        if group_id is None:
            self.edit_group_id.clear()
        else:
            self.edit_group_id.setText(str(group_id))
        items = self.labelList.findItems(text, QtCore.Qt.MatchFixedString)
        if items:
            if len(items) != 1:
                logger.warning("Label list has duplicate '{}'".format(text))
            self.labelList.setCurrentItem(items[0])
            row = self.labelList.row(items[0])
            self.edit.completer().setCurrentRow(row)
        self.edit.setFocus(QtCore.Qt.PopupFocusReason)
        if move:
            self.move(QtGui.QCursor.pos())
        if self.exec_():
            return self.edit.text(), self.getFlags(), self.getGroupId()
        else:
            return None, None, None

    def initLayout(self):
        """ 移除扩展的填选框 """
        layout = self.layout()
        for i in range(layout.rowCount() - 1, -1, -1):
            r = layout.itemAt(i, 1)
            if isinstance(r, QWidgetItem):
                w = r.widget()
                if w.property('labelme origin layout'):
                    w.show()
                else:
                    layout.removeRow(i)
        self.edit.setEnabled(True)

    def resetLayout(self, shape, mainwin):
        """ 增加布局项 """
        # 1 初始化
        self.initLayout()
        showtext, hashtext, labelattr = shape.parser(mainwin)
        if not labelattr:
            return
        layout = self.layout()
        cfg = mainwin.labelcfg

        # 2 处理labelattr的场景
        # 2.1 原有标准控件调整
        self.labelList.hide()
        self.edit.setEnabled(mainwin.label_editable.checked)

        # 2.2 封装添加每行控件的功能
        n_row = 1

        def add_row(k, v=None):
            nonlocal n_row

            def edit_label(k, x):
                """ 回调函数，控件修改值后，实时更新label """
                self.edit.setText(shape.setLabelAttr(self.edit.text(), k, x))
                self.setWindowModified(True)

            if k == 'label' and v and v['items'] is None and self.labelList.count():
                # label 字段特殊处理
                n = self.labelList.count()
                texts = [self.labelList.item(i).text() for i in range(n)]
                texts = [x for x in texts if (x and x[0] != '{' and x[-1] != '}')]
                v = v.copy()
                v['items'] = texts
                add_row(k, v) if texts else add_row(k)
            else:
                if v:
                    items, cvt = v['items'], v['type']
                else:
                    items, cvt = None, None

                cur_value = labelattr[k]
                if cur_value is None:
                    cur_value = 'None'
                w = get_input_widget(items, cur_value, valcvt=cvt,
                                     correct_changed=lambda x: edit_label(k, x))
                layout.insertRow(n_row, k, w)
                n_row += 1

        # 2.3 控件顺序
        # 2.3.1 控件分三组：明确不显示，明确要显示，未在设定清单中
        # labelattr中的所有字段都应该要展示，但要优先展示配置项里指定的，再展示其他次要、衍生字段
        keys_group = [[], [], []]
        for x in cfg.value:
            k, s = x['key'], x['show']
            if k in labelattr:
                keys_group[s].append(k)
        cfgkeys = cfg.keys()
        for k in labelattr.keys():
            if k not in cfgkeys:
                keys_group[2].append(k)
        # 2.3.2
        for g in [1, 2, 0]:  # 展示每组的顺序
            layout.insertRow(n_row, QHLine())  # 加分割线
            n_row += 1
            for k in keys_group[g]:
                add_row(k, cfg.get(k, None))
        layout.insertRow(n_row, QHLine())

    def popUp2(self, shape, mainwin=None):
        # 1 初始化
        text, flags, group_id = shape.label, shape.flags, shape.group_id
        if self._fit_to_content["row"]:
            self.labelList.setMinimumHeight(
                self.labelList.sizeHintForRow(0) * self.labelList.count() + 2
            )
        if self._fit_to_content["column"]:
            self.labelList.setMinimumWidth(
                self.labelList.sizeHintForColumn(0) + 2
            )
        # if text is None, the previous label in self.edit is kept
        if text is None:
            text = self.edit.text()
        if flags:
            self.setFlags(flags)
        else:
            self.resetFlags(text)
        self.edit.setText(text)
        self.edit.setSelection(0, len(text))
        if group_id is None:
            self.edit_group_id.clear()
        else:
            self.edit_group_id.setText(str(group_id))
        items = self.labelList.findItems(text, QtCore.Qt.MatchFixedString)
        if items:
            if len(items) != 1:
                logger.warning("Label list has duplicate '{}'".format(text))
            self.labelList.setCurrentItem(items[0])
            row = self.labelList.row(items[0])
            self.edit.completer().setCurrentRow(row)

        # 2 智能调整布局
        self.resetLayout(shape, mainwin)

        # 3 保存、返回shape结果
        if self.exec_():
            shape2 = ShapeExt()
            shape2.label = self.edit.text()
            shape2.flags = self.getFlags()
            shape2.group_id = self.getGroupId()
            return shape2
        else:
            return None
